from cesloi.delegatesystem import TemplateEvent
from pydantic.class_validators import validator

from cesloi.delegatesystem.entities.event import ParamsAnalysis
from .inserter import ApplicationInserter, EventInserter


class MiraiEvent(TemplateEvent):
    type: str

    @validator("type", allow_reuse=True)
    def type_limit(cls, v):
        if cls.type != v:
            raise Exception("{0}'s type must be '{1}', not '{2}'".format(cls.__name__, cls.type, v))
        return v

    class Config:
        extra = "ignore"


class EmptyEvent(TemplateEvent):
    def get_params(self, params):
        return ParamsAnalysis(
            ApplicationInserter,
            EventInserter
        ).error_param_check(params)



import inspect
from typing import Callable, Optional, Union
from pydantic import BaseModel
from inspect import iscoroutinefunction
from cesloi.message.messageChain import MessageChain
from cesloi.command import Command


class SubscriberInterface(BaseModel):
    callable_target: Callable

    class Config:
        arbitrary_types_allowed = True


class Subscriber(SubscriberInterface):
    def __init__(
            self,
            callable_target: Callable,
            subscriber_name: Optional[str] = None,
            command: Optional[Command] = None,
            require_param_name: str = None,
            is_replace_message: bool = True,
            time_schedule: Callable = None,
    ) -> None:
        super().__init__(
            callable_target=callable_target
        )
        self.command = command
        self.subscriber_name = subscriber_name or callable_target.__name__
        self.time_schedule = time_schedule
        self.require_param_name = require_param_name
        self.is_replace_message = is_replace_message
        if not iscoroutinefunction(callable_target):
            raise TypeError("Your Function must be a coroutine function (use async)!")

    def __call__(self, *args, **kwargs):
        return self.callable_target(*args, **kwargs)

    def __repr__(self):
        return f'<Subscriber,name={self.subscriber_name}>'

    def __str__(self):
        return self.__repr__()

    @property
    def name(self):
        return self.subscriber_name

    @name.setter
    def name(self, new_name):
        self.subscriber_name = new_name

    @property
    def params(self):
        return {
            name: [
                param.annotation if param.annotation is not inspect.Signature.empty else None,
                param.default if param.default is not inspect.Signature.empty else None,
            ]
            for name, param in inspect.signature(self.callable_target).parameters.items()
        }

    subscriber_name: Optional[str]
    command: Optional[Command] = None
    time_schedule: Callable = None
    require_param_name: str = None
    is_replace_message: bool = True


class SubscriberHandler:
    """
    订阅器相关的处理器
    """
    subscriber: Subscriber

    def set(
        self,
        subscriber_name: Optional[str] = None,
        command: Command = None,
        require_param_name: str = None,
        is_replace_message: bool = True
        # time_schedule: Callable = None,
    ):
        """该方法生成一个订阅器实例，该订阅器负责调控装饰的可执行函数

        若订阅器的注册事件是消息链相关的，可预先根据写入的命令相关参数进行判别是否执行该订阅器装饰的函数

        若不填如下参数则表示接受任意形式的命令

        Args:
            subscriber_name :装饰器名字，可选
            command :命令处理器，可选
            require_param_name :若命令处理器成功匹配文本并且非完全匹配模式(不含正则表达式)，则根据订阅器下的函数的参数中关于消息链参数的参数名判断是否将匹配到的文本通过该参数传递。
            is_replace_message : 此参数决定是否把匹配的文本传给消息链，默认为True

        """
        def wrapper(func):
            self.subscriber = Subscriber(
                func,
                subscriber_name,
                command,
                require_param_name,
                is_replace_message
                # time_schedule
            )
            return self.subscriber

        return wrapper

    @staticmethod
    def command_handler(sub: Subscriber, msg: Union[str, MessageChain]):
        if isinstance(msg, MessageChain):
            msg = msg.only_text()
        if not sub.command:
            return True, -1
        result = sub.command.analysis(msg)
        if result:
            return True, result

    @staticmethod
    def params_handler(sub: Subscriber, arg: dict, match_msg: str, msg: MessageChain):
        from ...message.element import Plain
        if sub.is_replace_message and match_msg != -1 and not isinstance(match_msg, bool):
            if sub.require_param_name:
                if isinstance(match_msg, str):
                    arg[sub.require_param_name] = msg.replace(Plain, Plain(match_msg), 1)
                if isinstance(match_msg, tuple):
                    arg[sub.require_param_name] = msg.replace(Plain, Plain(",".join(match_msg)), 1)
            else:
                for k, v in arg.items():
                    if v.__class__.__name__ == "MessageChain":
                        if isinstance(match_msg, str):
                            arg[k] = msg.replace(Plain, Plain(match_msg), 1)
                        if isinstance(match_msg, tuple):
                            arg[k] = msg.replace(Plain, Plain(",".join(match_msg)), 1)

        return arg



import loguru
import time


class Logger:
    logger: loguru.logger = loguru.logger

    def __init__(self, level='INFO'):
        self.logger.add('bot_running_log_' + time.strftime("%Y-%m-%d", time.localtime()) + '.log',
                        format="{time}|{level}|{message}", level=level, enqueue=True,
                        rotation="500MB", encoding="utf-8")
        self.logger.info("--------------------------------------------------------------------")



import os
from types import ModuleType
from typing import Optional, Dict, Union, Type, Callable
import importlib

from pydantic import BaseModel

from .delegatesystem import EventDelegate, TemplateEvent, Publisher
from .logger import Logger
from .command import Command


_module_target_dict: Dict[str, list] = {}


class CommandHandler(BaseModel):
    command: Command
    require_param_name: str = None,
    is_replace_message: bool = True

    def __init__(
            self,
            command: Command,
            require_param_name: str = None,
            is_replace_message: bool = True
    ):
        super().__init__(
            command=command,
            require_param_name=require_param_name,
            is_replace_message=is_replace_message
        )


class TemplatePlugin:
    module: ModuleType
    name: str
    usage: str
    is_close: bool = False

    def __init__(self, module, name:Optional[str] = None, usage: Optional[str] = None):
        if hasattr(module.__init__, "__annotations__"):
            module.__init__(**module.__init__.__annotations__)
        self.module = module
        self.name = name or ""
        self.usage = usage or ""


class Bellidin:
    """插件管理器

     - set_bellidin: 初始化管理器，通常不需要管
     - install_plugin: 载入单个模块，需要提供相对路径
     - install_plugins: 载入文件夹下的所有模块，需要提供相对路径
    """
    ignore = ["__init__.py", "__pycache__"]
    _modules: Dict[str, "TemplatePlugin"] = {}
    current_module_name: str
    delegate: EventDelegate
    logger: Logger.logger
    plugins_dir: str

    @classmethod
    def model_register(
            cls,
            event: Union[str, Type[TemplateEvent]],
            *,
            match_command: Optional[CommandHandler] = None
    ):
        if not cls.delegate:
            raise RuntimeError("Delegate didn't existed!")
        if isinstance(event, str):
            name = event
            event = cls.delegate.search_event(event)
            if not event:
                raise Exception(name + " cannot found!")

        def register_wrapper(func: Callable):
            from .delegatesystem.entities.subsciber import SubscriberHandler
            subscriber = SubscriberHandler().set(**match_command.dict() if match_command else None)(func)
            temp_publisher = cls.delegate.search_publisher(event)
            if not temp_publisher:
                cls.delegate.publisher_list.append(Publisher(subscriber, bind_event=event))
            else:
                if subscriber.name not in temp_publisher.subscribers_names():
                    temp_publisher.subscribers.append(subscriber)
                else:
                    raise ValueError(f"Function \"{subscriber.name}\" for {event.__name__} has already registered!")
            if cls.current_module_name not in _module_target_dict:
                _module_target_dict[cls.current_module_name] = [[event, subscriber]]
            else:
                _module_target_dict[cls.current_module_name].append([event, subscriber])
            return func

        return register_wrapper

    @classmethod
    def set_bellidin(cls, delegate, logger):
        cls.delegate = delegate
        cls.logger = logger
        cls.plugins_dir = ""
        return cls

    @classmethod
    def get_delegate(cls):
        return cls.delegate

    @classmethod
    def install_plugin(cls, modules_name: str):
        try:
            cls.current_module_name = modules_name
            module = importlib.import_module(modules_name, modules_name)
            name = getattr(module, '__name__', None)
            usage = getattr(module, '__usage', None)
            cls._modules[modules_name] = TemplatePlugin(module, name, usage)
            cls.logger.debug(f"plugin: {module.__name__} is installed")
            return True
        except Exception as e:
            cls.logger.exception(e)
            return False

    @classmethod
    def install_plugins(cls, plugin_dir: str):
        cls.plugins_dir = plugin_dir
        plugin_count = 0
        for module in os.listdir(plugin_dir):
            if module in cls.ignore:
                continue
            if os.path.isdir(module):
                cls.install_plugin(f"{plugin_dir.replace('/','.')}.{module}")
                plugin_count += 1
            else:
                cls.install_plugin(f"{plugin_dir.replace('/','.')}.{module.split('.')[0]}")
                plugin_count += 1
        cls.logger.info(f"{plugin_count} plugin have been installed")
        return plugin_count

    @classmethod
    def get_plugins(cls):
        return cls._modules

    @classmethod
    def uninstall_plugins(cls, *args, **kwargs):
        plugin_count = 0
        for module_name in cls._modules:
            targets = _module_target_dict[module_name]
            for target in targets:
                t_publisher = cls.delegate.search_publisher(target[0])
                t_publisher.remove_subscriber(target[1])
                if not t_publisher.subscribers:
                    cls.delegate.remove_publisher(t_publisher)
            _module_target_dict[module_name].clear()
            module = cls._modules[module_name].module
            if hasattr(module, "__end__"):
                try:
                    module.__end__(
                        *args, **module.__end__.__annotations__
                    )
                except Exception as e:
                    cls.logger.exception(e)
                    module.is_close = False
                else:
                    cls.logger.debug(f"plugin: {module.__name__} is uninstalled")
                    module.is_close = True
            plugin_count += 1
        cls._modules.clear()
        cls.logger.info(f"{plugin_count} plugins have been uninstalled successfully")
        return plugin_count

    @classmethod
    def uninstall_plugin(cls, module_name: str, *args, **kwargs):
        try:
            if module_name in cls._modules:
                targets = _module_target_dict[module_name]
                for target in targets:
                    t_publisher = cls.delegate.search_publisher(target[0])
                    t_publisher.remove_subscriber(target[1])
                    if not t_publisher.subscribers:
                        cls.delegate.remove_publisher(t_publisher)
                module = cls._modules[module_name].module
                if hasattr(module, "__end__"):
                    try:
                        module.__end__(
                            *args, **module.__end__.__annotations__
                        )
                    except Exception as e:
                        cls.logger.exception(e)
                        module.is_close = False
                    else:
                        cls.logger.debug(f"plugin: {module.__name__} is uninstalled")
                        module.is_close = True
            else:
                raise ValueError(f"No such plugin named {module_name}")
        except Exception as e:
            cls.logger.debug(e)
            pass

    @classmethod
    def reload_plugins(cls, new_plugins_dir: Optional[str] = None):
        try:
            if new_plugins_dir:
                cls.uninstall_plugins(cls.plugins_dir)
                cls.plugins_dir = new_plugins_dir
                cls.logger.info(f"reload plugins at {new_plugins_dir}")
                return cls.install_plugins(new_plugins_dir)
            else:
                cls.logger.info(f"reload plugins at {cls.plugins_dir}")
                plugin_count = 0
                for module_name in cls._modules:
                    targets = _module_target_dict[module_name]
                    for target in targets:
                        t_publisher = cls.delegate.search_publisher(target[0])
                        t_publisher.remove_subscriber(target[1])
                        if not t_publisher.subscribers:
                            cls.delegate.remove_publisher(t_publisher)
                    _module_target_dict[module_name].clear()
                    module = cls._modules[module_name].module
                    importlib.reload(module)
                    plugin_count += 1

                cls._modules.clear()
                cls.logger.info(f"{plugin_count} plugins have been reload successfully")
                return plugin_count
        except Exception as e:
            cls.logger.exception(e)
            pass



from cesloi.delegatesystem.entities.paramsInserter import TemplateParamsInserter
from cesloi.utils import bot_application, event


class ApplicationInserter(TemplateParamsInserter):
    name: str = "bot_application"

    @classmethod
    def get_param(cls):
        return bot_application.get()


class EventInserter(TemplateParamsInserter):
    name: str = "event"

    @classmethod
    def get_param(cls):
        return event.get()



from typing import Any, Dict, TYPE_CHECKING
from cesloi.delegatesystem.entities.event import ParamsAnalysis, TemplateEvent
if TYPE_CHECKING:
    from ..bot_client import Cesloi


class ApplicationRunning(TemplateEvent):
    bot: Any

    def __init__(self, bot):
        super().__init__(bot=bot)

    def get_params(self, params) -> Dict[str, Any]:
        return ParamsAnalysis(

        ).error_param_check(
            params,
            Cesloi=self.bot
        )


import functools
from contextvars import ContextVar
from contextlib import contextmanager
from enum import Enum
from typing import Callable, Any

bot_application = ContextVar("bot_application")
event = ContextVar("event")
event_loop = ContextVar("event_loop")
delegate = ContextVar("delegate")
upload_method = ContextVar("upload_method")


class ContextModel:
    content: Any

    @classmethod
    def set(cls, ct):
        cls.content = ct

    @classmethod
    def get(cls):
        return cls.content

    @classmethod
    def reset(cls, ct):
        cls.content = None
        cls.content = ct


class ContextCollection:
    bot_application: ContextModel
    event: ContextModel
    event_loop: ContextModel
    delegate: ContextModel
    upload_method: ContextModel



class UploadMethods(Enum):
    """用于向 `uploadImage` 或 `uploadVoice` 方法描述图片的上传类型"""

    Friend = "friend"
    Group = "group"
    Temp = "temp"


@contextmanager
def enter_message_send_context(method: UploadMethods):
    t = upload_method.set(method)
    yield
    upload_method.reset(t)


@contextmanager
def enter_context(bot=None, event_i=None):
    t1 = None
    t2 = None
    t3 = None
    t4 = None

    if bot:
        t1 = bot_application.set(bot)
        t3 = event_loop.set(bot.delegate.loop)
        t4 = delegate.set(bot.delegate)
    if event_i:
        t2 = event.set(event_i)

    yield
    try:
        if t1:
            bot_application.reset(t1)

        if all([t2, t3, t4]):
            event.reset(t2)
            event_loop.reset(t3)
            delegate.reset(t4)
    except ValueError:
        pass


def bot_application_context_manager(func: Callable):
    @functools.wraps(func)
    async def wrapper(self, *args, **kwargs):
        with enter_context(bot=self):
            return await func(self, *args, **kwargs)

    return wrapper



from typing import List, Type, Union

from pydantic import BaseModel

from .event import TemplateEvent
from .subsciber import Subscriber


class PublisherInterface(BaseModel):
    subscribers: List[Subscriber]
    bind_event: Type[TemplateEvent]


class Publisher(PublisherInterface):
    def __init__(
            self,
            *subscriber: List[Subscriber],
            bind_event: Type[TemplateEvent]
    ):
        super().__init__(
            subscribers=list(subscriber),
            bind_event=bind_event
        )

    def subscribers_names(self):
        return [i.name for i in self.subscribers]

    def remove_subscriber(self, target: Union[str, Subscriber]):
        if isinstance(target, Subscriber):
            self.subscribers.pop(self.subscribers.index(target))
        for i in range(0, len(self.subscribers)):
            if self.subscribers[i].name == target:
                self.subscribers.pop(i)
                break




from cesloi.event.base import MiraiEvent
from cesloi.delegatesystem.entities.event import ParamsAnalysis
from cesloi.message.messageChain import MessageChain
from .inserter import ApplicationInserter, EventInserter
from ..model.relation import Friend


class TestEvent(MiraiEvent):
    type: str = "TestEvent"
    messageChain: MessageChain
    sender: Friend

    def get_params(self, params):
        return ParamsAnalysis(
            ApplicationInserter,
            EventInserter
        ).error_param_check(
            params,
            MessageChain=self.messageChain,
            Friend=self.sender
        )


from pydantic import BaseModel


class TemplateParamsInserter(BaseModel):
    """
    参数插入模型
    """
    name: str

    @classmethod
    def get_param(cls):
        """该方法必须是 `classmethod`.

        该方法的返回值必须为实际元素数据.
        """
        pass



import asyncio
import traceback
import time
from typing import List, Union, Type, Dict, Any, Optional
from .entities.event import TemplateEvent
from .entities.publisher import Publisher
from .entities.subsciber import SubscriberHandler


class EventDelegate:
    loop: asyncio.AbstractEventLoop
    publisher_list: List[Publisher]
    safe_interval: float

    def __init__(self, loop: Optional[asyncio.AbstractEventLoop] = None, interval: Optional[float] = None):
        self.loop = loop or asyncio.get_event_loop()
        self.publisher_list = []
        self.safe_interval = interval or 0.001
        self.init_time = time.time()
        self.last_time = self.init_time

    def parse_to_event(self, event_dict: Dict[str, Any]):
        examine_event = self.search_event(event_dict.get('type')).parse_obj(
            {k: v for k, v in event_dict.items() if k != "type"})
        return examine_event

    def handle_event(self, event: TemplateEvent):
        current_time = time.time()
        if current_time - self.last_time > self.safe_interval:
            self.loop.create_task(
                self.exec_subscriber(
                    publisher=self.search_publisher(event.__class__),
                    event=event)
            )
        self.last_time = current_time

    async def exec_subscriber(self, publisher: Publisher, event: TemplateEvent):
        from ..event.messages import Message
        if publisher:
            coroutine = [
                self.get_coroutine_executable_target(
                    target,
                    event.get_params(target.params),  # 执行事件内的参数解析方法，获取可能的函数需要的实际参数
                    event.messageChain if isinstance(event, Message) else ""  # 获取可能的消息链
                ) for target in publisher.subscribers
            ]
            results, _ = await asyncio.wait(coroutine)
            for result in results:
                if result.exception() == "PropagationCancelled":
                    break

    @staticmethod
    async def get_coroutine_executable_target(target, real_arguments, message):
        # subscriber的command/time处理
        command_result = SubscriberHandler.command_handler(target, message)
        if command_result[0]:
            # 参数替换的处理
            real_arguments = SubscriberHandler.params_handler(target, real_arguments, command_result[1], message)
            try:
                result = await target.callable_target(**real_arguments)
            except Exception as e:
                traceback.print_exc()
                raise e
            return result

    def search_publisher(self, event: Type[TemplateEvent]):
        for publisher in self.publisher_list:
            if publisher.bind_event == event:
                return publisher
        return False

    def remove_publisher(self, target):
        self.publisher_list.remove(target)

    @staticmethod
    def event_class_generator(target=TemplateEvent):
        for i in target.__subclasses__():
            yield i
            if i.__subclasses__():
                yield from EventDelegate.event_class_generator(i)

    @staticmethod
    def search_event(name: str):
        for i in EventDelegate.event_class_generator():
            if i.__name__ == name:
                return i

    def register(self, event: Union[str, Type[TemplateEvent]]):
        if isinstance(event, str):
            name = event
            event = self.search_event(event)
            if not event:
                raise Exception(name + " cannot found!")

        def register_wrapper(subscriber):
            temp_publisher = self.search_publisher(event)
            if not temp_publisher:
                self.publisher_list.append(Publisher(subscriber, bind_event=event))
            else:
                if subscriber.name not in temp_publisher.subscribers_names():
                    temp_publisher.subscribers.append(subscriber)
                else:
                    raise ValueError(f"Function \"{subscriber.name}\" for {event.__name__} has already registered!")
            return subscriber

        return register_wrapper



from typing import Dict, List, Any
from pydantic.main import BaseModel
import functools
from typing import Callable
from .paramsInserter import TemplateParamsInserter


class TemplateEvent(BaseModel):
    class Config:
        arbitrary_types_allowed = True

    def get_params(self, params) -> Dict[str, Any]:
        """该方法可以是 `classmethod` 亦或是普通的方法/函数.

        唯一的要求是该方法的返回值必须为装有参数名—实际元素数据的字典.

        Args:
            params: 调用的函数的参数字典
        """
        return ParamsAnalysis().error_param_check(
            params,
        )


def extra_params_insert(func: Callable[[Dict], Dict]):
    """
    额外参数插入函数，会根据事件提供的额外参数插入器以及函数的参数列表进行相应的参数插入
    """
    @functools.wraps(func)
    def wrapper(self: "ParamsAnalysis", *args, **kwargs):
        for inserter in self.ParamsInserterList:
            param = inserter.get_param()
            if param.__class__ in list(map(lambda x: x[0], list(args[0].values()))):
                kwargs[param.__class__.__name__] = param
        return func(self, *args, **kwargs)
    return wrapper


class ParamsAnalysis:
    ParamsInserterList: List[TemplateParamsInserter]

    def __init__(self, *extra_param_inserter):
        """事件相关的参数分析模型

        Args:
            extra_param_inserter: 可提供的额外参数插入器,该插入器应继承自TemplateParamsInserter
        """
        self.ParamsInserterList = list(extra_param_inserter)

    @extra_params_insert
    def error_param_check(self, params, **kwargs):
        """
        将调用函数提供的参数字典与事件提供的参数字典进行比对，并返回正确的参数字典

        Args:
            params: 调用的函数的参数字典

            kwargs: 该事件可传入的参数字典,以” 类型 = 实际参数 “方式写入
        :return: 函数需要的参数字典
        """
        arguments_dict = {}
        for k, v in params.items():
            if v[0].__name__ in kwargs.keys():
                if isinstance(kwargs[v[0].__name__], Dict):
                    arguments_dict[k] = kwargs[v[0].__name__][v[1]]
                else:
                    arguments_dict[k] = kwargs[v[0].__name__]
            elif v[1] is not None:
                arguments_dict[k] = v[1]
        if len(params.items()) > len(arguments_dict.items()):
            raise ValueError(
                "Unexpected Extra Argument: \n" +
                "".join(map(
                    lambda x, y: "{0}:{1}, ".format(x, y[0].__name__)
                    if x not in arguments_dict else "", params.keys(), params.values()
                ))
            )
        return arguments_dict



import asyncio
import json
import random
from asyncio import Task

import aiohttp
from pydantic import BaseModel, AnyHttpUrl, Extra
from typing import Optional, Union, Dict, TYPE_CHECKING
from aiohttp import ClientSession, WSMsgType

from cesloi.utils import enter_context
from cesloi.delegatesystem import EventDelegate
from cesloi.logger import Logger
from yarl import URL

if TYPE_CHECKING:
    from .bot_client import Cesloi

code_exceptions_mapping = {
    1: "InvaildAuthkey  错误的verify key",
    2: "AccountNotFound  指定的Bot不存在",
    3: "InvaildSession  Session失效或不存在",
    4: "UnauthorizedSession  Session未认证(未激活)",
    5: "UnknownTarget  发送消息目标不存在(指定对象不存在)",
    6: "FileNotFoundError  指定文件不存在，出现于发送本地图片",
    10: "PermissionError  无操作权限，指Bot没有对应操作的限权",
    20: "AccountMuted  Bot被禁言，指Bot当前无法向指定群发送消息",
    30: "TooLongMessage   	消息过长",
    400: "InvaildArgument  错误的访问，如参数错误等",
}


def error_check(logger, code: Union[dict, int]):
    if isinstance(code, dict):
        code = code.get("code")
        exception_code = code_exceptions_mapping.get(code)
        if exception_code:
            logger.error(exception_code)
    elif isinstance(code, int):
        exception_code = code_exceptions_mapping.get(code)
        if exception_code:
            logger.error(exception_code)


class BotSession(BaseModel):
    host: AnyHttpUrl
    single_mode: bool = False
    account: Optional[int] = None
    verifyKey: Optional[str] = None
    sessionKey: Optional[str] = None
    version: Optional[str] = None

    def __init__(
            self,
            host: str,
            account: Optional[int] = None,
            verify_key: Optional[str] = None,
            *,
            single_mode: bool = False,
    ) -> None:
        """
            用于描述与上游接口会话, 并存储会话状态的实体类.

            Args:
                host (AnyHttpUrl): `mirai-api-http` 服务所在的根接口地址
                account (int): 应用所使用账号的整数 ID, singleMode 模式下为空, 非 singleMode 下新建连接不可为空.
                verify_key (str): 配置文件中指定,用以与mah绑定账号
            """
        super().__init__(
            host=host, account=account, verifyKey=verify_key, single_mode=single_mode
        )

    class Config:
        allow_mutation = True
        extra = Extra.allow


class Communicator:
    bot_session: BotSession
    delegate: EventDelegate
    loop: asyncio.AbstractEventLoop
    logger: Logger.logger

    def __init__(
            self,
            bot_session: BotSession,
            bot: "Cesloi",
            delegate: EventDelegate,
            logger: Optional[Logger] = None
    ):
        self.bot_session = bot_session
        self.delegate = delegate
        self.loop = delegate.loop
        self.logger = logger or Logger().logger
        self.bot = bot
        self.running_task: Optional[Task] = None
        self.running: bool = False
        self.ws_connection: Optional[aiohttp.ClientWebSocketResponse] = None
        self.client_session: Optional[ClientSession] = None
        self.wait_response_future: Dict[str, asyncio.Future] = {}

    async def stop(self):
        self.running = False
        if self.running_task and not self.running_task.done():
            try:
                await self.running_task
            except asyncio.CancelledError:
                pass
        self.running_task = None
        self.bot_session.sessionKey = None

    async def ws_send_handle(
            self,
            command_name: str, data: Optional[dict] = None, subcommand: Optional[str] = None):
        if not self.bot_session.verifyKey:
            raise ValueError
        sync_id = str(random.randint(0, 100_000_000))
        content = {
            'syncId': sync_id,
            'main': command_name,
            'content': json.dumps(data)
        }
        if subcommand:
            content['subcommand'] = subcommand
        await self.ws_connection.send_json(content)
        future = self.loop.create_future()
        self.wait_response_future[sync_id] = future
        result = await future
        del self.wait_response_future[sync_id]
        del future
        if "data" in result:
            return result['data']
        else:
            return result

    async def send_handle(
            self,
            action: str,
            method: str,
            data: Optional[dict] = None
    ):
        if not self.bot_session.verifyKey:
            raise ValueError
        data = data or dict()
        if method == "GET" or method == "get":
            async with self.client_session.get(
                    URL(f"{self.bot_session.host}/{action}").with_query(data)
            ) as response:
                response.raise_for_status()
                response_data = await response.json()
        elif method == "POST" or method == "update":
            async with self.client_session.post(
                    URL(f"{self.bot_session.host}/{action}"), data=json.dumps(data)
            ) as response:
                response.raise_for_status()
                response_data = await response.json()
        else:
            form = aiohttp.FormData()
            for k, v in data:
                form.add_fields(k, v)
            async with self.client_session.post(
                    URL(f"{self.bot_session.host}/{action}"), data=form
            ) as response:
                response.raise_for_status()
                response_data = await response.json()
        error_check(self.logger, response_data)
        if "data" in response_data:
            return response_data['data']
        else:
            return response_data

    async def ws_receive_handle(self, unknown_event_data: dict):
        if "syncId" in unknown_event_data:
            data, sync_id = unknown_event_data.get("data"), unknown_event_data.get("syncId")
            if sync_id == "-1":
                event = self.delegate.parse_to_event(data)
                with enter_context(bot=self.bot, event_i=event):
                    self.delegate.handle_event(event)
            else:
                if sync_id not in self.wait_response_future:
                    self.logger.warning(f"syncId {sync_id} not found!")
                else:
                    self.wait_response_future.pop(sync_id).set_result(data)

    async def receive_handle(self, unknown_event_data: dict):
        received_data = unknown_event_data.get('data')
        error_check(self.logger, received_data)
        event = self.delegate.parse_to_event(received_data)
        with enter_context(bot=self.bot, event_i=event):
            self.delegate.handle_event(event)

    async def websocket(self):
        """post_url = f"{self.bot_session.host}/all?" \
                   f"verifyKey={self.bot_session.verifyKey}".replace("http", "ws")
        if not self.bot_session.single_mode:
            post_url += f"&qq={self.bot_session.account}\""""
        async with self.client_session.ws_connect(
                str(URL(self.bot_session.host + "/all").with_query({"qq":self.bot_session.account,"verifyKey": self.bot_session.verifyKey})),
                autoping=False,
        ) as connection:
            self.logger.debug("connecting to websocket")
            self.ws_connection = connection
            connected = False
            ping_count = 0
            while self.running:
                try:
                    ws_message = await connection.receive()
                except asyncio.TimeoutError:
                    if ping_count > 5:
                        self.logger.warning("websocket: timeout,stop")
                        await self.stop()
                    else:
                        await self.ws_connection.ping()
                        ping_count += 1
                    continue
                if ws_message.type is WSMsgType.TEXT:
                    received_data: dict = json.loads(ws_message.data)
                    if connected:
                        try:
                            await self.receive_handle(received_data)
                        except Exception as e:
                            self.logger.exception(f"receive_data has error {e}")
                    else:
                        if not received_data['syncId']:
                            data = received_data['data']
                            if data['code']:
                                error_check(self.logger, data)

                            else:
                                if not self.bot_session.sessionKey:
                                    self.bot_session.sessionKey = data.get("session")
                                    connected = True
                elif ws_message.type is WSMsgType.CLOSE:
                    self.logger.info("websocket: server close connection.")
                    return
                elif ws_message.type is WSMsgType.CLOSED:
                    self.logger.info("websocket: connection has been closed.")
                    return
                elif ws_message.type is WSMsgType.PONG:
                    self.logger.debug("websocket: received pong from remote")
                elif ws_message.type == WSMsgType.ERROR:
                    self.logger.warning("websocket: connection error: " + ws_message.data)
                else:
                    self.logger.warning(f"detected a unknown message type: {ws_message.type}")
        self.logger.info("connection disconnected")

    async def connect(self):
        if not self.client_session:
            self.client_session = ClientSession()
        if not self.running_task or self.running_task.done():
            self.running = True
            self.running_task = self.loop.create_task(self.websocket())
